# Generated by Django 4.1.10 on 2023-07-24 15:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0043_alter_instrumentlistthroughmodel_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="instrumentlist",
            name="identifier",
            field=models.CharField(blank=True, max_length=255, unique=True),
        ),
        migrations.AddField(
            model_name="instrumentlistthroughmodel",
            name="computed_str",
            field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
        ),
        migrations.RenameIndex(
            model_name="instrumentprice",
            new_name="instrument_price_base_idx",
            old_name="instrument_price_idx",
        ),
        migrations.AddIndex(
            model_name="instrumentprice",
            index=models.Index(fields=["calculated", "instrument"], name="instrument_price_idx1"),
        ),
        migrations.AddIndex(
            model_name="instrumentprice",
            index=models.Index(fields=["instrument"], name="instrument_price_idx2"),
        ),
    ]
