# Generated by Django 4.2.9 on 2024-02-02 10:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbportfolio", "0050_fees_fee_date_fees_wbportfolio_transac_1f7a29_idx"),
    ]

    operations = []
